from __future__ import annotations

import asyncio
import json
import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import TypeVar

import websockets.client
import websockets.exceptions
from exponent.commands.utils import ConnectionTracker
from exponent.core.remote_execution import files, git, system_context
from exponent.core.remote_execution.code_execution import execute_code
from exponent.core.remote_execution.command_execution import execute_command
from exponent.core.remote_execution.file_write import execute_file_write
from exponent.core.remote_execution.session import (
    RemoteExecutionClientSession,
    get_session,
)
from exponent.core.remote_execution.types import (
    CLIConnectedState,
    CodeExecutionRequest,
    CommandRequest,
    CreateChatResponse,
    ExecutionEndResponse,
    FileWriteRequest,
    GetAllTrackedFilesRequest,
    GetFileAttachmentRequest,
    GetFileAttachmentsRequest,
    GetMatchingFilesRequest,
    HeartbeatInfo,
    ListFilesRequest,
    RemoteExecutionRequestType,
    RemoteExecutionResponseType,
    StartChatRequest,
    StartChatResponse,
    SystemContextRequest,
    UseToolsConfig,
)
from exponent.core.remote_execution.utils import (
    assert_unreachable,
    deserialize_api_response,
    deserialize_request_data,
    serialize_message,
)
from httpx import (
    AsyncClient,
    codes as http_status,
)
from pydantic import BaseModel

logger = logging.getLogger(__name__)


TModel = TypeVar("TModel", bound=BaseModel)


class RemoteExecutionClient:
    def __init__(
        self,
        session: RemoteExecutionClientSession,
    ):
        self.current_session = session

        self.file_cache: files.FileCache = files.FileCache(session.working_directory)

    @property
    def working_directory(self) -> str:
        return self.current_session.working_directory

    @property
    def api_client(self) -> AsyncClient:
        return self.current_session.api_client

    async def run_connection(  # noqa: PLR0915,PLR0912
        self,
        chat_uuid: str,
        connection_tracker: ConnectionTracker | None = None,
    ) -> None:
        stop = False

        async for websocket in self.ws_connect(
            f"/api/ws/chat/{chat_uuid}",
        ):
            if connection_tracker is not None:
                await connection_tracker.set_connected(True)

            beats: asyncio.Queue[HeartbeatInfo] = asyncio.Queue()
            requests: asyncio.Queue[RemoteExecutionRequestType] = asyncio.Queue()
            results: asyncio.Queue[RemoteExecutionResponseType] = asyncio.Queue()

            async def beat() -> None:
                while True:
                    info = self.get_heartbeat_info()
                    await beats.put(info)
                    await asyncio.sleep(3)

            async def executor() -> None:
                while True:
                    request = await requests.get()
                    response = await self.handle_request(request)
                    await results.put(response)

            beat_task = asyncio.create_task(beat())
            executor_task = asyncio.create_task(executor())

            try:
                while True:
                    recv = asyncio.create_task(websocket.recv())
                    get_beat = asyncio.create_task(beats.get())
                    get_result = asyncio.create_task(results.get())

                    done, pending = await asyncio.wait(
                        [recv, get_beat, get_result],
                        return_when=asyncio.FIRST_COMPLETED,
                    )

                    for task in pending:
                        task.cancel()

                        try:
                            await task
                        except asyncio.CancelledError:
                            pass

                    if recv in done:
                        msg = json.loads(str(recv.result()))

                        if msg["type"] == "request":
                            data = json.dumps(msg["data"])
                            request = deserialize_request_data(data)
                            await requests.put(request)

                    elif get_beat in done:
                        info = get_beat.result()
                        data = json.loads(info.model_dump_json())
                        msg = json.dumps({"type": "heartbeat", "data": data})
                        await websocket.send(msg)

                    elif get_result in done:
                        response = get_result.result()
                        data = json.loads(serialize_message(response))
                        msg = json.dumps({"type": "result", "data": data})
                        await websocket.send(msg)

            except websockets.exceptions.ConnectionClosed as e:
                if e.rcvd is not None and e.rcvd.code == 1000:  # noqa: PLR2004
                    stop = True
            except TimeoutError:
                pass

            beat_task.cancel()
            executor_task.cancel()

            if stop:
                break

            if connection_tracker is not None:
                await connection_tracker.set_connected(False)

    async def check_remote_end_event(self, chat_uuid: str) -> bool:
        response = await self.api_client.get(
            f"/api/remote_execution/{chat_uuid}/execution_end",
        )
        execution_end_response = await deserialize_api_response(
            response, ExecutionEndResponse
        )
        return execution_end_response.execution_ended

    async def create_chat(self) -> CreateChatResponse:
        response = await self.api_client.post(
            "/api/remote_execution/create_chat",
        )
        return await deserialize_api_response(response, CreateChatResponse)

    async def start_chat(
        self, chat_uuid: str, prompt: str, use_tools_config: UseToolsConfig
    ) -> StartChatResponse:
        response = await self.api_client.post(
            "/api/remote_execution/start_chat",
            json=StartChatRequest(
                chat_uuid=chat_uuid,
                prompt=prompt,
                use_tools_config=use_tools_config,
            ).model_dump(),
            timeout=60,
        )
        return await deserialize_api_response(response, StartChatResponse)

    def get_heartbeat_info(self) -> HeartbeatInfo:
        return HeartbeatInfo(
            system_info=system_context.get_system_info(self.working_directory),
        )

    async def send_heartbeat(self, chat_uuid: str) -> CLIConnectedState:
        logger.info(f"Sending heartbeat for chat_uuid {chat_uuid}")
        heartbeat_info = self.get_heartbeat_info()
        response = await self.api_client.post(
            f"/api/remote_execution/{chat_uuid}/heartbeat",
            content=heartbeat_info.model_dump_json(),
            timeout=60,
        )
        if response.status_code != http_status.OK:
            raise Exception(
                f"Heartbeat failed with status code {response.status_code} and response {response.text}"
            )
        connected_state = await deserialize_api_response(response, CLIConnectedState)
        logger.info(f"Heartbeat response: {connected_state}")
        return connected_state

    async def handle_request(
        self, request: RemoteExecutionRequestType
    ) -> RemoteExecutionResponseType:
        response: RemoteExecutionResponseType
        if isinstance(request, CodeExecutionRequest):
            response = await execute_code(
                request, self.current_session, working_directory=self.working_directory
            )
        elif isinstance(request, FileWriteRequest):
            response = execute_file_write(
                request, working_directory=self.working_directory
            )
        elif isinstance(request, ListFilesRequest):
            response = files.list_files(request)
        elif isinstance(request, GetFileAttachmentRequest):
            response = files.get_file_attachment(request)
        elif isinstance(request, GetFileAttachmentsRequest):
            response = files.get_file_attachments(request)
        elif isinstance(request, GetMatchingFilesRequest):
            response = await files.get_matching_files(request, self.file_cache)
        elif isinstance(request, SystemContextRequest):
            response = system_context.get_system_context(
                request, self.working_directory
            )
        elif isinstance(request, GetAllTrackedFilesRequest):
            response = await git.get_all_tracked_files(request, self.working_directory)
        elif isinstance(request, CommandRequest):
            response = await execute_command(request, self.working_directory)
        else:
            assert_unreachable(request)
        return response

    def ws_connect(self, path: str) -> websockets.client.connect:
        base_url = (
            str(self.api_client.base_url)
            .replace("http://", "ws://")
            .replace("https://", "wss://")
        )

        url = f"{base_url}{path}"
        headers = {"api-key": self.api_client.headers["api-key"]}

        return websockets.client.connect(
            url, extra_headers=headers, timeout=10, ping_timeout=10
        )

    @staticmethod
    @asynccontextmanager
    async def session(
        api_key: str, base_url: str, working_directory: str
    ) -> AsyncGenerator[RemoteExecutionClient, None]:
        async with get_session(working_directory, base_url, api_key) as session:
            yield RemoteExecutionClient(session)



import pygit2

from exponent.core.remote_execution.files import normalize_files
from exponent.core.remote_execution.git import (
    get_all_tracked_files,
    get_all_tracked_git_file_paths,
    get_git_info,
    get_git_subtree_for_dir,
)
from exponent.core.remote_execution.types import (
    GetAllTrackedFilesRequest,
    GetAllTrackedFilesResponse,
    GitInfo,
)


async def test_empty_git_repo(temporary_directory_no_commit_history: str) -> None:
    expected_files = normalize_files(
        temporary_directory_no_commit_history, ["test1.py", "test2.py", "exponent.txt"]
    )

    repo = pygit2.Repository(temporary_directory_no_commit_history)
    subtree = get_git_subtree_for_dir(repo, temporary_directory_no_commit_history)
    assert subtree is None

    tracked_files_response = await get_all_tracked_files(
        request=GetAllTrackedFilesRequest(
            correlation_id="123",
        ),
        working_directory=temporary_directory_no_commit_history,
    )
    assert tracked_files_response == GetAllTrackedFilesResponse(
        correlation_id="123",
        files=expected_files,
    )

    git_info = get_git_info(working_directory=temporary_directory_no_commit_history)
    assert git_info == GitInfo(
        branch="master",
        remote=None,
    )


async def test_non_empty_git_repo(default_temporary_directory: str) -> None:
    expected_files = normalize_files(
        default_temporary_directory, ["test1.py", "test2.py", "exponent.txt"]
    )

    repo = pygit2.Repository(default_temporary_directory)
    tracked_files = get_all_tracked_git_file_paths(repo, default_temporary_directory)
    assert sorted(tracked_files) == expected_files

    tracked_files_response = await get_all_tracked_files(
        request=GetAllTrackedFilesRequest(
            correlation_id="123",
        ),
        working_directory=default_temporary_directory,
    )
    assert tracked_files_response == GetAllTrackedFilesResponse(
        correlation_id="123",
        files=expected_files,
    )

    git_info = get_git_info(working_directory=default_temporary_directory)
    assert git_info == GitInfo(
        branch="master",
        remote=None,
    )


async def test_non_git_repo(temporary_directory_no_git: str) -> None:
    expected_files = normalize_files(
        temporary_directory_no_git,
        ["test1.py", "test2.py", "exponent.txt", "node_modules_not_ignored/test3.py"],
    )

    tracked_files_response = await get_all_tracked_files(
        request=GetAllTrackedFilesRequest(
            correlation_id="123",
        ),
        working_directory=temporary_directory_no_git,
    )
    assert tracked_files_response == GetAllTrackedFilesResponse(
        correlation_id="123",
        files=expected_files,
    )

    git_info = get_git_info(working_directory=temporary_directory_no_git)
    assert git_info is None



import asyncio
import os
import subprocess
import sys
import threading
import time
from json import JSONDecodeError

import click
import httpx
from exponent.core.config import Environment, ExponentCloudConfig, Settings
from exponent.version import get_version
from packaging.version import Version


def print_editable_install_forced_prod_warning(settings: Settings) -> None:
    click.secho(
        "Detected local editable install, but this command only works against prod.",
        fg="red",
        bold=True,
    )
    click.secho("Using prod settings:", fg="red", bold=True)
    click.secho("- base_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_url}", fg=(100, 200, 255), bold=False)
    click.secho("- base_api_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_api_url}", fg=(100, 200, 255), bold=False)
    click.secho()


def print_editable_install_warning(settings: Settings) -> None:
    click.secho(
        "Detected local editable install, using local URLs", fg="yellow", bold=True
    )
    click.secho("- base_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_url}", fg=(100, 200, 255), bold=False)
    click.secho("- base_api_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_api_url}", fg=(100, 200, 255), bold=False)
    click.secho()


def print_exponent_message(base_url: str, chat_uuid: str) -> None:
    click.echo()
    click.secho("△ Exponent v1.0.0", fg=(180, 150, 255), bold=True)
    click.echo()
    click.echo(
        " - Link: " + click.style(f"{base_url}/chats/{chat_uuid}", fg=(100, 200, 255))
    )
    click.echo(click.style("  - Shell: /bin/zsh", fg="white"))
    click.echo()
    click.echo(click.style("✓", fg="green", bold=True) + " Ready in 1401ms")


def is_exponent_app_installed() -> bool:
    if sys.platform == "darwin":  # macOS
        return os.path.exists("/Applications/Exponent.app")

    # TODO: Add support for Windows and Linux
    return False


def launch_exponent_browser(
    environment: Environment, base_url: str, chat_uuid: str
) -> None:
    if is_exponent_app_installed() and environment == Environment.production:
        url = f"exponent://chats/{chat_uuid}"
    else:
        url = f"{base_url}/chats/{chat_uuid}"
    click.launch(url)


def write_template_exponent_cloud_config(file_path: str) -> None:
    exponent_cloud_config = ExponentCloudConfig(
        repo_name="your_repo_name",
        repo_specific_setup_commands=[
            "cd /home/user",
            "gh repo clone https://github.com/<org>/<repo>.git",
            "cd <repo>",
            "# Any additional setup commands",
        ],
        gh_token="ghp_your_token_here",
        runloop_api_key="ak_your_runloop_api_key_here",
    )
    with open(file_path, "w") as f:
        f.write(exponent_cloud_config.model_dump_json(indent=2))


def get_exponent_version_for_update() -> str | None:
    """
    Check if there's a newer version of Exponent available on PyPI.
    """
    installed_version = get_version()

    try:
        pypi_package_info = httpx.get("https://pypi.org/pypi/exponent-run/json").json()
    except (httpx.ConnectError, JSONDecodeError):
        return None
    pypi_version = pypi_package_info.get("info", {}).get("version")

    if not pypi_version or installed_version == "unknown":
        # If we can't get the concrete version info,
        # we can't tell if it's outdated
        return None
    elif Version(pypi_version) > Version(installed_version):
        return str(pypi_version)
    return None


def check_exponent_version() -> None:
    """
    Print a message to the user if there's a newer
    version of Exponent available on PyPI.
    """
    current_version = get_version()
    if current_version == "unknown":
        click.secho("Unable to determine current Exponent version.", fg="yellow")
        return

    new_version = get_exponent_version_for_update()
    if not new_version:
        return

    upgrade_command = ["pip", "install", "--upgrade", "exponent-run"]
    click.secho(
        f"New version available: exponent-run=={new_version} (current: {current_version})\n"
        f"Update command: `{' '.join(upgrade_command)}`",
        fg="yellow",
        bold=True,
    )
    confirmed = click.confirm("Update now?", default=True, show_default=True)
    if not confirmed:
        click.secho("Not updating.", fg="red")
        return

    upgrade_exponent(new_version)


def upgrade_exponent(new_version: str) -> None:
    """
    Upgrade Exponent to the latest version.
    """
    upgrade_command = ["pip", "install", "--upgrade", "exponent-run"]
    click.secho("Updating...", bold=True, fg="yellow")
    subprocess.check_call(upgrade_command)
    click.secho(f"Successfully upgraded Exponent to version {new_version}!", fg="green")
    invoking_command = subprocess.list2cmdline(["exponent", *sys.argv[1:]])
    click.secho(
        "\nUpdate complete! Please re-run command: ",
        fg="green",
        nl=False,
    )
    click.secho(
        invoking_command,
        bold=True,
    )
    sys.exit(0)


def start_background_event_loop() -> asyncio.AbstractEventLoop:
    def run_event_loop(loop: asyncio.AbstractEventLoop) -> None:
        asyncio.set_event_loop(loop)
        loop.run_forever()

    loop = asyncio.new_event_loop()
    thread = threading.Thread(target=run_event_loop, args=(loop,), daemon=True)
    thread.start()
    return loop


def read_input(prompt: str) -> str:
    sys.stdout.write(prompt)
    sys.stdout.flush()
    return sys.stdin.readline()


class Spinner:
    def __init__(self, text: str) -> None:
        self.text = text
        self.task: asyncio.Task[None] | None = None
        self.base_time = time.time()

    def show(self) -> None:
        if self.task is not None:
            return

        async def spinner(base_time: float) -> None:
            chars = "⣷⣯⣟⡿⢿⣻⣽⣾"

            while True:
                t = time.time() - base_time
                i = round(t * 10) % len(chars)
                print(f"\r{chars[i]} {self.text}", end="")
                await asyncio.sleep(0.1)

        self.task = asyncio.get_event_loop().create_task(spinner(self.base_time))

    def hide(self) -> None:
        if self.task is None:
            return

        self.task.cancel()
        self.task = None
        print("\r\x1b[0m\x1b[2K", end="")
        sys.stdout.flush()


class ConnectionTracker:
    def __init__(self) -> None:
        self.connected = True
        self.queue: asyncio.Queue[bool] = asyncio.Queue()

    def is_connected(self) -> bool:
        while True:
            try:
                self.connected = self.queue.get_nowait()
            except asyncio.QueueEmpty:
                break

        return self.connected

    async def wait_for_reconnection(self) -> None:
        if not self.is_connected():
            assert await self.queue.get()
            self.connected = True

    async def set_connected(self, connected: bool) -> None:
        await self.queue.put(connected)

    async def next_change(self) -> bool:
        return await self.queue.get()



import os

from exponent.core.remote_execution.command_execution import execute_command
from exponent.core.remote_execution.file_write import (
    execute_search_replace_edit,
    execute_udiff_edit,
)
from exponent.core.remote_execution.languages.shell import execute_shell
from exponent.core.remote_execution.types import CommandRequest, CommandResponse
from exponent.core.types.generated.command_request_data import (
    FileReadCommandRequestData,
)


async def test_execute_shell(default_temporary_directory: str) -> None:
    output = await execute_shell(
        code="ls -1v",
        working_directory=default_temporary_directory,
        timeout=5,
    )

    output_lines = output.strip().split("\n")
    expected_lines = 6

    assert len(output_lines) == expected_lines

    assert output_lines == [
        "exponent.txt",
        "symlink.txt",
        "test1.py",
        "test2.py",
        "",
        "EXIT CODE: 0",
    ]

    output = await execute_shell(
        code="true && echo 'hi' || echo \"boo\"",
        working_directory=default_temporary_directory,
        timeout=5,
    )

    assert output.strip() == "hi\n\nEXIT CODE: 0"


async def test_execute_udiff(temporary_directory_with_bigger_file: str) -> None:
    diff = (
        "@@ ... @@\n"
        " def subtract(a: int, b: int) -> int:\n"
        "-    return a ++ b\n"
        "+    return a - b\n"
        " \n"
        " def add(a: int, b: int) -> int:\n"
    )
    execute_udiff_edit(
        file_path="test1.py",
        content=diff,
        working_directory=temporary_directory_with_bigger_file,
    )

    contents = None
    with open(os.path.join(temporary_directory_with_bigger_file, "test1.py")) as f:
        contents = f.read()

    assert (
        contents
        == """
def subtract(a: int, b: int) -> int:
    return a - b

def add(a: int, b: int) -> int:
    return a + b

def multiply(a: int, b: int) -> int:
    return a * b
"""
    )


async def test_execute_udiff_fuzzy(temporary_directory_with_bigger_file: str) -> None:
    diff = (
        "@@ ... @@\n"
        " def subtract(a: int, b: int) -> int:\n"
        "-    return a ++ b\n"
        "+    return a - b\n"
        " def add(a: int, b: int) -> int:\n"
    )
    execute_udiff_edit(
        file_path="test1.py",
        content=diff,
        working_directory=temporary_directory_with_bigger_file,
    )

    contents = None
    with open(os.path.join(temporary_directory_with_bigger_file, "test1.py")) as f:
        contents = f.read()

    assert (
        contents
        == """
def subtract(a: int, b: int) -> int:
    return a - b

def add(a: int, b: int) -> int:
    return a + b

def multiply(a: int, b: int) -> int:
    return a * b
"""
    )


async def test_execute_search_replace(
    temporary_directory_with_bigger_file: str,
) -> None:
    diff = (
        "<<<<< SEARCH\n"
        "def subtract(a: int, b: int) -> int:\n"
        "    return a ++ b\n"
        "======\n"
        "def subtract(a: int, b: int) -> int:\n"
        "    return a - b\n"
        ">>>>>> REPLACE\n"
    )
    execute_search_replace_edit(
        file_path="test1.py",
        content=diff,
        working_directory=temporary_directory_with_bigger_file,
    )

    contents = None
    with open(os.path.join(temporary_directory_with_bigger_file, "test1.py")) as f:
        contents = f.read()

    assert (
        contents
        == """
def subtract(a: int, b: int) -> int:
    return a - b

def add(a: int, b: int) -> int:
    return a + b

def multiply(a: int, b: int) -> int:
    return a * b
"""
    )


async def test_execute_search_replace_fuzzy(
    temporary_directory_with_bigger_file: str,
) -> None:
    diff = (
        "<<<<SEARCH\n"
        "def subtract(a: int, b: int) -> int:\n"
        "    return a ++ b\n"
        "======\n"
        "def subtract(a: int, b: int) -> int:\n"
        "    return a - b\n"
        ">>>>>>>>>> REPLACE\n"
    )
    execute_search_replace_edit(
        file_path="test1.py",
        content=diff,
        working_directory=temporary_directory_with_bigger_file,
    )

    contents = None
    with open(os.path.join(temporary_directory_with_bigger_file, "test1.py")) as f:
        contents = f.read()

    assert (
        contents
        == """
def subtract(a: int, b: int) -> int:
    return a - b

def add(a: int, b: int) -> int:
    return a + b

def multiply(a: int, b: int) -> int:
    return a * b
"""
    )


async def test_execute_command(default_temporary_directory: str) -> None:
    request = CommandRequest(
        correlation_id="123456",
        data=FileReadCommandRequestData(file_path="test1.py", language="python"),
    )

    response = await execute_command(
        request, working_directory=default_temporary_directory
    )

    assert response == CommandResponse(
        content="print('Hello, world!')",
        correlation_id="123456",
    )



GET_CHATS_QUERY: str = """
    query Chats {
        chats {
            ... on UnauthenticatedError {
                message
            }
            ... on Chats {
                chats {
                    id
                    chatUuid
                    name
                    firstLine
                    isShared
                    isStarted
                    updatedAt # ISO datetime string
                }
            }
        }
    }
"""



import os
from typing import cast

import pygit2


def initialize_repo(path: str) -> pygit2.Repository:
    if not os.path.exists(path):
        os.makedirs(path)
    return cast(pygit2.Repository, pygit2.init_repository(path, bare=False))


def stage_file(repo: pygit2.Repository, filepath: str) -> None:
    index = repo.index
    index.add(filepath)
    index.write()


def create_commit(
    repo: pygit2.Repository, message: str, author_name: str, author_email: str
) -> None:
    # Retrieve the index and check if it has changes
    index = repo.index
    if not index.write_tree():
        print("No changes to commit.")
        return

    # Get the tree object
    tree = index.write_tree()

    # Get the current HEAD. It could be a new repo without commits.
    try:
        parents = [repo.head.target]
    except pygit2.GitError:
        parents = []

    # Create the author and committer signatures
    author = pygit2.Signature(author_name, author_email)

    # Do the commit
    repo.create_commit(
        "refs/heads/master",  # the name of the reference to update
        author,  # the author of the commit
        author,  # the committer of the commit
        message,  # the commit message
        tree,  # the tree object this commit points to
        parents,  # parents of this commit (empty list for the first commit)
    )



import getpass
import os
import platform
from pathlib import Path
from unittest import mock
from unittest.mock import mock_open

import pytest
from click.testing import CliRunner

from exponent.commands.config_commands import login
from exponent.core.remote_execution import files, git
from exponent.core.remote_execution.client import RemoteExecutionClient
from exponent.core.remote_execution.languages.python import Kernel, get_python_env_info
from exponent.core.remote_execution.system_context import get_system_context
from exponent.core.remote_execution.types import (
    FileWriteRequest,
    FileWriteResponse,
    GetAllTrackedFilesRequest,
    GetAllTrackedFilesResponse,
    GetFileAttachmentRequest,
    GetFileAttachmentResponse,
    GetMatchingFilesRequest,
    GetMatchingFilesResponse,
    GitInfo,
    ListFilesRequest,
    PythonEnvInfo,
    RemoteFile,
    SystemContextRequest,
    SystemContextResponse,
    SystemInfo,
)


@pytest.mark.skip
async def test_exponent_login(cli_runner: CliRunner) -> None:
    with mock.patch("builtins.open", new_callable=mock_open):
        result = cli_runner.invoke(
            login, ["--key", "123456"], env={"EXPONENT_API_KEY": "123456"}
        )
        assert result.exit_code == 0
        assert "Saving API Key" in result.output


async def test_list_files(default_temporary_directory: str) -> None:
    request = ListFilesRequest(
        correlation_id="123456", directory=default_temporary_directory
    )

    response = files.list_files(request)

    assert response.correlation_id == "123456"
    assert sorted(response.files) == [
        RemoteFile(
            file_path="exponent.txt",
            working_directory=default_temporary_directory,
        ),
        RemoteFile(
            file_path="symlink.txt",
            working_directory=default_temporary_directory,
        ),
        RemoteFile(
            file_path="test1.py",
            working_directory=default_temporary_directory,
        ),
        RemoteFile(
            file_path="test2.py",
            working_directory=default_temporary_directory,
        ),
        RemoteFile(
            file_path=".git",
            working_directory=default_temporary_directory,
        ),
    ]


async def test_get_system_context(
    default_temporary_directory: str, python_env_info: PythonEnvInfo
) -> None:
    request = SystemContextRequest(correlation_id="123456")

    response = get_system_context(request, default_temporary_directory)

    assert response == SystemContextResponse(
        correlation_id="123456",
        exponent_txt="Hello, world!",
        system_info=SystemInfo(
            name=getpass.getuser(),
            cwd=default_temporary_directory,
            shell=os.environ.get("SHELL", "bash"),
            os=platform.system(),
            git=GitInfo(branch="master", remote=None),
            python_env=python_env_info,
        ),
    )


async def test_get_file_attachment(default_temporary_directory: str) -> None:
    request = GetFileAttachmentRequest(
        correlation_id="123456",
        file=RemoteFile(
            file_path="test1.py",
            working_directory=default_temporary_directory,
        ),
    )

    response = files.get_file_attachment(request)

    assert response == GetFileAttachmentResponse(
        correlation_id="123456",
        file=RemoteFile(
            file_path="test1.py",
            working_directory=default_temporary_directory,
        ),
        content="print('Hello, world!')",
    )


async def test_get_matching_files(default_temporary_directory: str) -> None:
    request1 = GetMatchingFilesRequest(correlation_id="123456", search_term="tes")
    request2 = GetMatchingFilesRequest(correlation_id="123456", search_term="test1")

    response1 = await files.get_matching_files(
        request1, files.FileCache(default_temporary_directory)
    )

    assert response1.correlation_id == "123456"

    assert sorted(response1.files) == [
        RemoteFile(
            file_path="exponent.txt",
            working_directory=default_temporary_directory,
        ),
        RemoteFile(
            file_path="test1.py",
            working_directory=default_temporary_directory,
        ),
        RemoteFile(
            file_path="test2.py",
            working_directory=default_temporary_directory,
        ),
    ]

    response2 = await files.get_matching_files(
        request2, files.FileCache(default_temporary_directory)
    )

    assert response2.correlation_id == "123456"

    assert sorted(response2.files) == [
        RemoteFile(
            file_path="exponent.txt",
            working_directory=default_temporary_directory,
        ),
        RemoteFile(
            file_path="test1.py",
            working_directory=default_temporary_directory,
        ),
        RemoteFile(
            file_path="test2.py",
            working_directory=default_temporary_directory,
        ),
    ]


async def test_get_matching_files_nested(temporary_directory: str) -> None:
    # Setup test directory to look like this:
    # /folder
    #   /subfolder
    #     test1.py
    #     test2.py
    # /main
    #   core.java
    # random.tsx
    # .env
    # .gitignore (contains .env ignored)

    # Test queries for:
    # .env
    # random
    # test1
    # test2
    # .tsx
    # mjava
    import os

    # Create the necessary files and folders in the temporary directory
    os.makedirs(os.path.join(temporary_directory, "folder", "subfolder"))
    os.makedirs(os.path.join(temporary_directory, "main"))

    with open(
        os.path.join(temporary_directory, "folder", "subfolder", "test1.py"),
        "w",
    ) as f:
        f.write("# test1.py")

    with open(
        os.path.join(temporary_directory, "folder", "subfolder", "test2.py"),
        "w",
    ) as f:
        f.write("# test2.py")

    with open(os.path.join(temporary_directory, "main", "core.java"), "w") as f:
        f.write("// core.java")

    with open(os.path.join(temporary_directory, "random.tsx"), "w") as f:
        f.write("// random.tsx")

    with open(os.path.join(temporary_directory, ".env"), "w") as f:
        f.write("SECRET_KEY=123456")

    with open(os.path.join(temporary_directory, ".gitignore"), "w") as f:
        f.write(".env")

    # Test queries for various search terms
    test_cases = [
        # ("tes", ["folder/subfolder/test1.py", "folder/subfolder/test2.py"]),
        ("random", ["random.tsx"]),
        (".env", []),
        ("core", ["main/core.java"]),
        (".tsx", ["random.tsx"]),
        ("java", ["main/core.java"]),
        ("nonexistent", []),
        ("folder", ["folder/subfolder/test1.py", "folder/subfolder/test2.py"]),
        ("subfolder", ["folder/subfolder/test1.py", "folder/subfolder/test2.py"]),
        ("test1", ["folder/subfolder/test1.py"]),
        ("test2", ["folder/subfolder/test2.py"]),
    ]

    for search_term, _ in test_cases:
        request = GetMatchingFilesRequest(
            correlation_id="123456", search_term=search_term
        )

        response = await files.get_matching_files(
            request, files.FileCache(temporary_directory)
        )

        assert isinstance(response, GetMatchingFilesResponse)
        assert response.correlation_id == "123456"


@pytest.mark.skip
async def test_python_kernel() -> None:
    kernel = Kernel(os.getcwd())
    result = await kernel.execute_code("print('Hello, world!')")
    assert result == "Hello, world!\n"


async def test_execute_code_diff(tmpdir: Path) -> None:
    os.chdir(tmpdir)

    request = FileWriteRequest(
        correlation_id="123456",
        file_path="echo_server.py",
        language="python",
        write_strategy="FULL_FILE_REWRITE",
        content='from fastapi import FastAPI\nfrom pydantic import BaseModel\n\napp = FastAPI()\n\nclass Message(BaseModel):\n    content: str\n\n@app.get("/echo")\ndef echo(message: Message):\n    return {"Echo": message.content}\n\nif __name__ == "__main__":\n    import uvicorn\n    uvicorn.run(app, host="127.0.0.1", port=8000)\n',
    )

    async with RemoteExecutionClient.session(
        api_key="123456",
        base_url="https://example.com",
        working_directory=str(tmpdir),
    ) as client:
        response = await client.handle_request(request)

        assert response == FileWriteResponse(
            correlation_id="123456", content="Created file echo_server.py"
        )

        with open("echo_server.py") as f:
            contents = f.read()
        assert (
            contents
            == 'from fastapi import FastAPI\nfrom pydantic import BaseModel\n\napp = FastAPI()\n\nclass Message(BaseModel):\n    content: str\n\n@app.get("/echo")\ndef echo(message: Message):\n    return {"Echo": message.content}\n\nif __name__ == "__main__":\n    import uvicorn\n    uvicorn.run(app, host="127.0.0.1", port=8000)\n'
        )


async def test_get_all_tracked_files(temporary_directory_with_folder: str) -> None:
    request = GetAllTrackedFilesRequest(correlation_id="123456")

    response = await git.get_all_tracked_files(request, temporary_directory_with_folder)

    assert response == GetAllTrackedFilesResponse(
        correlation_id="123456",
        files=[
            RemoteFile(
                file_path="exponent.txt",
                working_directory=temporary_directory_with_folder,
            ),
            RemoteFile(
                file_path="test1.py",
                working_directory=temporary_directory_with_folder,
            ),
            RemoteFile(
                file_path="test2.py",
                working_directory=temporary_directory_with_folder,
            ),
            RemoteFile(
                file_path="nested/exponent.txt",
                working_directory=temporary_directory_with_folder,
            ),
            RemoteFile(
                file_path="nested/test1.py",
                working_directory=temporary_directory_with_folder,
            ),
            RemoteFile(
                file_path="nested/test2.py",
                working_directory=temporary_directory_with_folder,
            ),
        ],
    )

    response = await git.get_all_tracked_files(
        request, f"{temporary_directory_with_folder}/nested"
    )

    assert response == GetAllTrackedFilesResponse(
        correlation_id="123456",
        files=[
            RemoteFile(
                file_path="nested/exponent.txt",
                working_directory=temporary_directory_with_folder,
            ),
            RemoteFile(
                file_path="nested/test1.py",
                working_directory=temporary_directory_with_folder,
            ),
            RemoteFile(
                file_path="nested/test2.py",
                working_directory=temporary_directory_with_folder,
            ),
        ],
    )


def test_get_python_env_info() -> None:
    venv_info = get_python_env_info()
    assert venv_info.interpreter_path
    assert venv_info.interpreter_version
    assert venv_info.interpreter_path.endswith("python")
    assert venv_info.interpreter_version == platform.python_version()



import asyncio
import os
import tempfile
from collections.abc import Generator
from pathlib import Path
from unittest.mock import patch

import pytest
from click.testing import CliRunner

from exponent.core.remote_execution.languages import python
from exponent.core.remote_execution.types import PythonEnvInfo
from exponent.tests.utils import create_commit, initialize_repo, stage_file


@pytest.fixture(scope="session")
def event_loop() -> Generator[asyncio.AbstractEventLoop, None, None]:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.close()


@pytest.fixture
def cli_runner() -> CliRunner:
    return CliRunner()


@pytest.fixture(scope="session", autouse=True)
async def fixture_setup_env() -> None:
    session_mpatch = pytest.MonkeyPatch()
    session_mpatch.setenv("EXPONENT_API_KEY", "123456")
    session_mpatch.setenv("EXPONENT_BASE_URL", "https://exponent.run")
    session_mpatch.setenv("EXPONENT_API_BASE_URL", "https://api.exponent.run")


@pytest.fixture(scope="function")
def temporary_directory() -> Generator[str, None, None]:
    with tempfile.TemporaryDirectory() as test_directory:
        yield test_directory


@pytest.fixture(scope="function")
def temporary_directory_2() -> Generator[str, None, None]:
    with tempfile.TemporaryDirectory() as test_directory:
        yield test_directory


@pytest.fixture(scope="function")
def default_temporary_directory(
    temporary_directory: str, temporary_directory_2: str
) -> Generator[str, None, None]:
    with open(os.path.join(temporary_directory, "test1.py"), "w") as f:
        f.write("print('Hello, world!')")

    with open(os.path.join(temporary_directory, "test2.py"), "w") as f:
        f.write("print('Hello, world!')")

    with open(os.path.join(temporary_directory, "exponent.txt"), "w") as f:
        f.write("Hello, world!")

    symlink_path = os.path.join(temporary_directory_2, "symlink.txt")
    with open(symlink_path, "w") as f:
        f.write("Hello, world!")
    os.symlink(symlink_path, os.path.join(temporary_directory, "symlink.txt"))

    repo = initialize_repo(temporary_directory)

    stage_file(repo, "test1.py")
    stage_file(repo, "test2.py")
    stage_file(repo, "exponent.txt")
    create_commit(repo, "Initial commit", "Tester McTesterface", "test@example.com")

    yield temporary_directory


@pytest.fixture(scope="function")
def temporary_directory_no_commit_history(
    temporary_directory: str,
) -> Generator[str, None, None]:
    with open(os.path.join(temporary_directory, "test1.py"), "w") as f:
        f.write("print('Hello, world!')")

    with open(os.path.join(temporary_directory, "test2.py"), "w") as f:
        f.write("print('Hello, world!')")

    with open(os.path.join(temporary_directory, "exponent.txt"), "w") as f:
        f.write("Hello, world!")

    initialize_repo(temporary_directory)

    yield temporary_directory


@pytest.fixture(scope="function")
def temporary_directory_no_git(
    temporary_directory: str,
) -> Generator[str, None, None]:
    base_path = Path(temporary_directory)

    test1py = base_path / "test1.py"
    test1py.write_text("print('Hello, world!')")

    test2py = base_path / "test2.py"
    test2py.write_text("print('Hello, world!')")

    exptxt = base_path / "exponent.txt"
    exptxt.write_text("Hello, world!")

    file_in_ignored_pattern = Path(temporary_directory) / "test.pyc"
    file_in_ignored_pattern.write_text("Hello, world!")

    ignored_dir = Path(temporary_directory) / "node_modules"
    ignored_dir.mkdir(parents=True, exist_ok=True)

    file_in_ignored_dir = Path(temporary_directory) / "node_modules" / "nested.js"
    file_in_ignored_dir.write_text('console.log("Hello, world!");')

    not_ignored_dir = Path(temporary_directory) / "node_modules_not_ignored"
    not_ignored_dir.mkdir(parents=True, exist_ok=True)

    file_in_not_ignored_dir = (
        Path(temporary_directory) / "node_modules_not_ignored" / "test3.py"
    )
    file_in_not_ignored_dir.write_text('print("Hello, world!");')

    yield temporary_directory


@pytest.fixture(scope="function")
def temporary_directory_with_folder(
    temporary_directory: str,
) -> Generator[str, None, None]:
    base_path = Path(temporary_directory)

    test1py = base_path / "test1.py"
    test1py.write_text("print('Hello, world!')")

    test2py = base_path / "test2.py"
    test2py.write_text("print('Hello, world!')")

    exptxt = base_path / "exponent.txt"
    exptxt.write_text("Hello, world!")

    nested_path = Path(temporary_directory) / "nested"
    nested_path.mkdir(parents=True, exist_ok=True)

    nested_test1py = nested_path / "test1.py"
    nested_test1py.write_text("print('Hello, nested world!')")

    nested_test2py = nested_path / "test2.py"
    nested_test2py.write_text("print('Hello, nested world!')")

    nested_exptxt = nested_path / "exponent.txt"
    nested_exptxt.write_text("Hello, nested world!")

    repo = initialize_repo(temporary_directory)

    stage_file(repo, "test1.py")
    stage_file(repo, "test2.py")
    stage_file(repo, "exponent.txt")
    stage_file(repo, "nested/test1.py")
    stage_file(repo, "nested/test2.py")
    stage_file(repo, "nested/exponent.txt")
    create_commit(repo, "Initial commit", "Tester McTesterface", "test@example.com")

    yield temporary_directory


@pytest.fixture(scope="function", autouse=True)
def mock_cli_heartbeat() -> Generator[None, None, None]:
    patcher = patch(
        "exponent.core.remote_execution.client.RemoteExecutionClient.send_heartbeat"
    )
    patcher.start()
    yield
    patcher.stop()


@pytest.fixture(name="python_env_info", scope="function", autouse=True)
def fixture_python_env_info(
    monkeypatch: pytest.MonkeyPatch,
) -> PythonEnvInfo:
    env_info = PythonEnvInfo(
        interpreter_path="python",
        interpreter_version="3.11.0",
    )

    def get_python_env_info() -> PythonEnvInfo:
        return env_info

    monkeypatch.setattr(
        python,
        "get_python_env_info",
        get_python_env_info,
    )

    return env_info


@pytest.fixture(scope="function")
def temporary_directory_with_bigger_file(
    temporary_directory: str,
) -> Generator[str, None, None]:
    with open(os.path.join(temporary_directory, "test1.py"), "w") as f:
        f.write("""
def subtract(a: int, b: int) -> int:
    return a ++ b

def add(a: int, b: int) -> int:
    return a + b

def multiply(a: int, b: int) -> int:
    return a * b
""")

    repo = initialize_repo(temporary_directory)

    stage_file(repo, "test1.py")
    create_commit(repo, "Initial commit", "Tester McTesterface", "test@example.com")

    yield temporary_directory



import asyncio
import shutil
import sys
from collections.abc import Callable, Coroutine, Iterable
from concurrent.futures import Future
from typing import Any

import click
from exponent.commands.common import (
    create_chat,
    inside_ssh_session,
    redirect_to_login,
    start_client,
)
from exponent.commands.settings import use_settings
from exponent.commands.types import (
    StrategyChoice,
    StrategyOption,
    exponent_cli_group,
)
from exponent.commands.utils import (
    ConnectionTracker,
    Spinner,
    start_background_event_loop,
)
from exponent.core.config import Settings
from exponent.core.graphql.client import GraphQLClient
from exponent.core.graphql.mutations import HALT_CHAT_STREAM_MUTATION
from exponent.core.graphql.subscriptions import CHAT_EVENTS_SUBSCRIPTION
from exponent.core.remote_execution.exceptions import ExponentError
from exponent.core.types.generated.strategy_info import (
    ENABLED_STRATEGY_INFO_LIST,
)
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.key_binding import KeyBindings
from rich.console import Console
from rich.syntax import Syntax

SLASH_COMMANDS = {
    "/help": "Show available commands",
    "/autorun": "Toggle autorun mode",
    "/web": "Move chat to a web browser",
}


@exponent_cli_group()
def shell_cli() -> None:
    pass


@shell_cli.command()
@click.option(
    "--model",
    help="LLM model",
    required=True,
    default="CLAUDE_3_POINT_5_SONNET",
)
@click.option(
    "--strategy",
    prompt=True,
    prompt_required=False,
    type=StrategyChoice(ENABLED_STRATEGY_INFO_LIST),
    cls=StrategyOption,
)
@click.option(
    "--autorun",
    is_flag=True,
    help="Enable autorun mode",
)
@click.option(
    "--prompt",
    help="Initial prompt",
)
@use_settings
def shell(
    settings: Settings,
    model: str,
    strategy: str,
    autorun: bool = False,
    prompt: str | None = None,
) -> None:
    """Start an Exponent session in your current shell."""

    if not settings.api_key:
        redirect_to_login(settings)
        return

    api_key = settings.api_key
    base_api_url = settings.base_api_url
    gql_client = GraphQLClient(api_key, base_api_url)
    loop = start_background_event_loop()

    chat_uuid = asyncio.run_coroutine_threadsafe(
        create_chat(api_key, base_api_url), loop
    ).result()

    if chat_uuid is None:
        sys.exit(1)

    connection_tracker = ConnectionTracker()

    client_fut = asyncio.run_coroutine_threadsafe(
        start_client(
            api_key, base_api_url, chat_uuid, connection_tracker=connection_tracker
        ),
        loop,
    )

    input_handler = InputHandler()
    chat = Chat(chat_uuid, settings.base_url, gql_client, model, strategy, autorun)
    shell = Shell(prompt, loop, input_handler, chat, connection_tracker)
    shell.run()
    client_fut.cancel()
    print("\rBye!")


def pause_spinner(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(self: "Chat", *args: Any, **kwargs: Any) -> Any:
        self.spinner.hide()
        result = func(self, *args, **kwargs)
        self.spinner.show()
        return result

    return wrapper


def stop_spinner(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(self: "Chat", *args: Any, **kwargs: Any) -> Any:
        self.spinner.hide()
        result = func(self, *args, **kwargs)
        return result

    return wrapper


class Chat:
    def __init__(  # noqa: PLR0913
        self,
        chat_uuid: str,
        base_url: str,
        gql_client: GraphQLClient,
        model: str,
        strategy: str,
        autorun: bool,
    ) -> None:
        self.chat_uuid = chat_uuid
        self.base_url = base_url
        self.gql_client = gql_client
        self.model = model
        self.strategy = strategy
        self.autorun = autorun
        self.parent_uuid: str | None = None
        self.block_row_offset = 0
        self.console = Console()
        self.code_block_uuid = None
        self.file_write_uuid = None
        self.command_uuid = None
        self.block: Block = NullBlock()
        self.default_spinner = Spinner("Exponent is working...")
        self.exec_spinner = Spinner(
            "Exponent is waiting for the code to finish running..."
        )
        self.spinner = self.default_spinner

    async def send_prompt(self, prompt: str) -> str | None:
        self.spinner.show()

        return await self.process_chat_subscription(
            {"prompt": {"message": prompt, "attachments": []}}
        )

    async def send_confirmation(self) -> str | None:
        if self.code_block_uuid is not None:
            return await self.process_chat_subscription(
                {
                    "codeBlockConfirmation": {
                        "accepted": True,
                        "codeBlockUuid": self.code_block_uuid,
                    }
                }
            )
        elif self.file_write_uuid is not None:
            return await self.process_chat_subscription(
                {
                    "fileWriteConfirmation": {
                        "accepted": True,
                        "fileWriteUuid": self.file_write_uuid,
                    }
                }
            )
        elif self.command_uuid is not None:
            return await self.process_chat_subscription(
                {
                    "commandConfirmation": {
                        "accepted": True,
                        "commandUuid": self.command_uuid,
                    }
                }
            )

        return None

    def toggle_autorun(self) -> bool:
        self.autorun = not self.autorun
        return self.autorun

    async def halt_stream(self) -> None:
        await self.gql_client.query(
            HALT_CHAT_STREAM_MUTATION, {"chatUuid": self.chat_uuid}, "HaltChatStream"
        )

    def url(self) -> str:
        return f"{self.base_url}/chats/{self.chat_uuid}"

    async def process_chat_subscription(  # noqa: PLR0912, PLR0915
        self,
        extra_vars: dict[str, Any],
    ) -> str | None:
        result = None

        vars = {
            "chatUuid": self.chat_uuid,
            "parentUuid": self.parent_uuid,
            "model": self.model,
            "strategyNameOverride": self.strategy,
            "useToolsConfig": "read_write",
            "requireConfirmation": not self.autorun,
        }

        vars.update(extra_vars)
        self.code_block_uuid = None
        self.file_write_uuid = None
        self.command_uuid = None

        async for response in self.gql_client.subscribe(CHAT_EVENTS_SUBSCRIPTION, vars):
            event = response["authenticatedChat"]
            kind = event["__typename"]

            self.parent_uuid = event.get("eventUuid") or self.parent_uuid

            if kind == "MessageChunkEvent":
                if event["role"] == "assistant":
                    self.handle_message_chunk_event(event)
            elif kind == "MessageEvent":
                if event["role"] == "assistant":
                    self.handle_message_event(event)
            elif kind == "FileWriteChunkEvent":
                self.handle_file_write_chunk_event(event)
            elif kind == "FileWriteEvent":
                self.handle_file_write_event(event)
            elif kind == "FileWriteConfirmationEvent":
                self.handle_file_write_confirmation_event(event)
            elif kind == "FileWriteStartEvent":
                self.handle_file_write_start_event(event)
            elif kind == "FileWriteResultEvent":
                self.handle_file_write_result_event(event)
            elif kind == "CodeBlockChunkEvent":
                self.handle_code_block_chunk_event(event)
            elif kind == "CodeBlockEvent":
                self.handle_code_block_event(event)
            elif kind == "CodeBlockConfirmationEvent":
                self.handle_code_block_confirmation_event(event)
            elif kind == "CodeExecutionStartEvent":
                self.handle_code_execution_start_event(event)
            elif kind == "CodeExecutionEvent":
                self.handle_code_execution_event(event)
            elif kind == "CommandChunkEvent":
                if event["data"]["type"] not in ["FILE_READ", "FILE_OPEN"]:
                    continue
                self.handle_command_chunk_event(event)
            elif kind == "CommandEvent":
                if event["data"]["type"] not in ["FILE_READ", "FILE_OPEN"]:
                    continue
                self.handle_command_event(event)
            elif kind == "CommandConfirmationEvent":
                self.handle_command_confirmation_event(event)
            elif kind == "CommandStartEvent":
                self.handle_command_start_event(event)
            elif kind == "CommandResultEvent":
                self.handle_command_result_event(event)
            elif kind == "Error":
                raise ExponentError(event["message"])
            else:
                # TODO
                pass

        self.spinner.hide()
        return result

    @stop_spinner
    def handle_message_chunk_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs = []

        if self.block.event_uuid != event_uuid:
            self.block = MessageBlock(event_uuid)

        assert isinstance(self.block, MessageBlock)

        seqs.append(self.block.render_message(event["content"]))
        render(seqs)

    @pause_spinner
    def handle_message_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs = []

        if self.block.event_uuid != event_uuid:
            self.block = MessageBlock(event_uuid)

        assert isinstance(self.block, MessageBlock)

        seqs.append(self.block.render_message(event["content"]))
        seqs.append(["\n\n"])
        render(seqs)

    @stop_spinner
    def handle_file_write_chunk_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs = []

        if self.block.event_uuid != event_uuid:
            self.block = FileWriteBlock(event_uuid, event["language"])

            seqs.append(
                [
                    self.render_block_header(f"Editing file {event['filePath']}"),
                    self.render_block_padding(),
                ]
            )

        assert isinstance(self.block, FileWriteBlock)

        seqs.append(self.block.render_content(event["content"]))
        render(seqs)

    @pause_spinner
    def handle_file_write_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs: list[Any] = []

        if self.block.event_uuid != event_uuid:
            self.block = FileWriteBlock(event_uuid, event["language"])

            seqs.append(
                [
                    self.render_block_header(f"Editing file {event['filePath']}"),
                    self.render_block_padding(),
                ]
            )

        assert isinstance(self.block, FileWriteBlock)

        seqs.append(self.block.render_content(event["content"]))
        seqs.append(self.render_block_padding())

        if event["requireConfirmation"]:
            self.file_write_uuid = event_uuid

            seqs.append(
                [
                    self.render_block_footer(
                        "Confirm edit with <C+y>. Sending a new message will dismiss code changes."
                    ),
                    "\n",
                ]
            )

        render(seqs)

    @pause_spinner
    def handle_file_write_confirmation_event(self, event: dict[str, Any]) -> None:
        if self.block.event_uuid != event["fileWriteUuid"]:
            return

        assert isinstance(self.block, FileWriteBlock)

        seqs = []

        if event["accepted"]:
            seqs.append(
                [
                    move_cursor_up_seq(2),
                    self.render_block_footer("⚙️ Applying edit..."),
                    "\n",
                ]
            )
        else:
            # user entered new prompt, cursor moved down
            # therefore we need to move it up, redraw status, and move it
            # back where it was

            seqs.append(
                [
                    move_cursor_up_seq(4),
                    self.render_block_footer("✘ Edit dismissed"),
                    move_cursor_down_seq(3),
                ]
            )

        render(seqs)

    @pause_spinner
    def handle_file_write_start_event(self, event: dict[str, Any]) -> None:
        return

    @pause_spinner
    def handle_file_write_result_event(self, event: dict[str, Any]) -> None:
        if self.block.event_uuid != event["fileWriteUuid"]:
            return

        assert isinstance(self.block, FileWriteBlock)

        seqs = []
        result = event["content"]

        seqs.append(
            [
                move_cursor_up_seq(2),
                self.render_block_footer(f"✔️ {result}" if result else "✔️ Edit applied"),
                move_cursor_down_seq(1),
            ]
        )

        render(seqs)

    @stop_spinner
    def handle_command_chunk_event(self, event: dict[str, Any]) -> None:
        type_ = event["data"]["type"].lower()
        event_uuid = event["eventUuid"]
        seqs = []

        if self.block.event_uuid != event_uuid:
            self.block = CommandBlock(event_uuid, type_, event["data"]["language"])

            seqs.append(
                [
                    self.render_block_header(type_),
                    self.render_block_padding(),
                ]
            )

        assert isinstance(self.block, CommandBlock)

        seqs.append(self.block.render_data(event["data"]))
        render(seqs)

    @pause_spinner
    def handle_command_event(self, event: dict[str, Any]) -> None:
        type_ = event["data"]["type"].lower()
        event_uuid = event["eventUuid"]
        seqs: list[Any] = []

        if self.block.event_uuid != event_uuid:
            self.block = CommandBlock(event_uuid, type_, event["data"]["language"])

            seqs.append(
                [
                    self.render_block_header(type_),
                    self.render_block_padding(),
                ]
            )

        assert isinstance(self.block, CommandBlock)

        seqs.append(self.block.render_data(event["data"]))
        seqs.append(self.render_block_padding())

        if event["requireConfirmation"]:
            self.command_uuid = event_uuid

            seqs.append(
                [
                    self.render_block_footer(
                        f"Confirm {type_} command with <C+y>. "
                        "Sending a new message will cancel this command."
                    ),
                    "\n",
                ]
            )

        render(seqs)

    @pause_spinner
    def handle_command_confirmation_event(self, event: dict[str, Any]) -> None:
        if self.block.event_uuid != event["commandUuid"]:
            return

        assert isinstance(self.block, CommandBlock)

        seqs = []

        if event["accepted"]:
            seqs.append(
                [
                    move_cursor_up_seq(2),
                    self.render_block_footer(
                        f"⚙️ Executing {self.block.type} command..."
                    ),
                    "\n",
                ]
            )
        else:
            # user entered new prompt, cursor moved down
            # therefore we need to move it up, redraw status, and move it
            # back where it was

            seqs.append(
                [
                    move_cursor_up_seq(4),
                    self.render_block_footer(
                        f"✘ {self.block.type} command did not execute"
                    ),
                    move_cursor_down_seq(3),
                ]
            )

        render(seqs)

    @pause_spinner
    def handle_command_start_event(self, event: dict[str, Any]) -> None:
        return

    @pause_spinner
    def handle_command_result_event(self, event: dict[str, Any]) -> None:
        if self.block.event_uuid != event["commandUuid"]:
            return

        assert isinstance(self.block, CommandBlock)

        seqs = []
        result = event["content"]
        result = pad_left(result.strip(), "  ")

        seqs.append(
            [
                move_cursor_up_seq(2),
                self.render_block_header("output"),
                self.render_block_padding(),
                self.block.highlight_code(result, self.block.lang, line_numbers=False),
                self.render_block_padding(),
                self.render_block_footer(f"✔️ {self.block.type} command executed"),
                move_cursor_down_seq(1),
            ]
        )

        render(seqs)

    @stop_spinner
    def handle_code_block_chunk_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs = []

        if self.block.event_uuid != event_uuid:
            self.block = CodeBlock(event_uuid, event["language"])

            seqs.append(
                [
                    self.render_block_header(event["language"]),
                    self.render_block_padding(),
                ]
            )

        assert isinstance(self.block, CodeBlock)

        seqs.append(self.block.render_code(event["content"]))
        render(seqs)

    @pause_spinner
    def handle_code_block_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs: list[Any] = []

        if self.block.event_uuid != event_uuid:
            self.block = CodeBlock(event_uuid, event["language"])

            seqs.append(
                [
                    self.render_block_header(event["language"]),
                    self.render_block_padding(),
                ]
            )

        assert isinstance(self.block, CodeBlock)

        seqs.append(self.block.render_code(event["content"]))
        seqs.append(self.render_block_padding())

        if event["requireConfirmation"]:
            self.code_block_uuid = event_uuid

            seqs.append(
                [
                    self.render_block_footer(
                        "Run this code now with <C+y>. Sending a new message will cancel this request."
                    ),
                    "\n",
                ]
            )

        render(seqs)

    @pause_spinner
    def handle_code_block_confirmation_event(self, event: dict[str, Any]) -> None:
        if self.block.event_uuid != event["codeBlockUuid"]:
            return

        assert isinstance(self.block, CodeBlock)

        seqs = []

        if event["accepted"]:
            seqs.append(
                [
                    move_cursor_up_seq(2),
                    self.render_block_footer("⚙️ Running code..."),
                    "\n",
                ]
            )
        else:
            # user entered new prompt, cursor moved down
            # therefore we need to move it up, redraw status, and move it
            # back where it was

            seqs.append(
                [
                    move_cursor_up_seq(4),
                    self.render_block_footer("✘ Code did not execute"),
                    move_cursor_down_seq(3),
                ]
            )

        render(seqs)

    @pause_spinner
    def handle_code_execution_start_event(self, event: dict[str, Any]) -> None:
        self.spinner = self.exec_spinner

    @pause_spinner
    def handle_code_execution_event(self, event: dict[str, Any]) -> None:
        self.spinner = self.default_spinner

        if self.block.event_uuid != event["codeBlockUuid"]:
            return

        assert isinstance(self.block, CodeBlock)

        seqs = []
        lines = pad_left(event["content"].strip(), "  ").split("\n")
        output = [[block_body_bg_seq(), erase_line_seq(), line, "\n"] for line in lines]

        seqs.append(
            [
                move_cursor_up_seq(2),
                self.render_block_header("output"),
                self.render_block_padding(),
                output,
                self.render_block_padding(),
                move_cursor_down_seq(1),
            ]
        )

        render(seqs)

    def render_block_header(self, text: str) -> list[str]:
        return [
            block_header_bg_seq(),
            block_header_fg_seq(),
            erase_line_seq(),
            " " + text,
            reset_attrs_seq(),
            "\n",
        ]

    def render_block_padding(self) -> list[str]:
        return [
            block_body_bg_seq(),
            erase_line_seq(),
            reset_attrs_seq(),
            "\n",
        ]

    def render_block_footer(self, text: str) -> list[str]:
        return [
            block_header_bg_seq(),
            block_header_fg_seq(),
            erase_line_seq(),
            " " + text,
            reset_attrs_seq(),
            "\n",
        ]


class InputHandler:
    def __init__(self) -> None:
        self.kb = KeyBindings()
        self.shortcut_pressed = None
        self.session: PromptSession[Any] = PromptSession(
            completer=SlashCommandCompleter(),
            complete_while_typing=False,
            key_bindings=self.kb,
        )

        @self.kb.add("c-y")
        def _(event: Any) -> None:
            self.shortcut_pressed = "<c-y>"
            event.app.exit()

        @self.kb.add("c-d")
        def _(event: Any) -> None:
            self.shortcut_pressed = "<c-d>"
            event.app.exit()

    def prompt(self) -> str:
        self.shortcut_pressed = None
        user_input = self.session.prompt(HTML("<b><ansigreen>></ansigreen></b> "))

        if self.shortcut_pressed:
            return self.shortcut_pressed
        else:
            assert isinstance(user_input, str)
            return user_input


class Shell:
    def __init__(  # noqa PLR0913
        self,
        prompt: str | None,
        loop: asyncio.AbstractEventLoop,
        input_handler: InputHandler,
        chat: Chat,
        connection_tracker: ConnectionTracker,
    ) -> None:
        self.prompt = prompt
        self.loop = loop
        self.input_handler = input_handler
        self.chat = chat
        self.stream_fut: Future[Any] | None = None
        self.connection_tracker = connection_tracker

    def run(self) -> None:
        self._print_welcome_message()
        self._send_initial_prompt()

        while True:
            try:
                self._wait_for_stream_completion()
                text = self.input_handler.prompt()

                if text.startswith("/"):
                    self._run_command(text[1:].strip())
                elif text == "<c-y>":
                    self._confirm_execution()
                elif text in {"q", "exit"}:
                    print()
                    break
                elif text == "<c-d>":
                    print()
                    do_quit = self._ask_for_quit_confirmation()
                    print()

                    if do_quit:
                        break
                elif text:
                    print()
                    self._send_prompt(text)

            except KeyboardInterrupt:
                if self.stream_fut is not None:
                    self._run_coroutine(self.chat.halt_stream()).result()
                else:
                    do_quit = self._ask_for_quit_confirmation()
                    print()

                    if do_quit:
                        break

            except ExponentError as e:
                self._print_error_message(e)
                break

    def _ensure_connected(self) -> None:
        if not self.connection_tracker.is_connected():
            self._run_coroutine(self._wait_for_reconnection()).result()

    async def _wait_for_reconnection(self) -> None:
        render([clear_line_seq(), "Disconnected..."])
        await asyncio.sleep(1)
        spinner = Spinner("Reconnecting...")
        spinner.show()
        await self.connection_tracker.wait_for_reconnection()
        spinner.hide()
        render([fg_color_seq(2), "\x1b[1m", "✓ Reconnected"])
        await asyncio.sleep(1)
        render([clear_line_seq()])

    def _print_welcome_message(self) -> None:
        print("Welcome to ✨ \x1b[1;32mExponent \x1b[4:3mSHELL\x1b[0m ✨")
        print()
        print("Type 'q', 'exit' or press <C-c> to exit")
        print("Enter '/help' to see a list of available commands")
        print()

    def _print_error_message(self, e: ExponentError) -> None:
        print(f"\n\n\x1b[1;31m{e}\x1b[0m")
        print("\x1b[3;33m")
        print("Please try again and reach out if the problem persists.")
        print("\x1b[0m")

    def _show_help(self) -> None:
        print()

        for command, description in SLASH_COMMANDS.items():
            print(f"{command} - {description}")

        print()

    def _run_command(self, command: str) -> None:
        if command == "help":
            self._show_help()
        elif command == "autorun":
            self._toggle_autorun()
        elif command == "web":
            self._switch_chat_to_web()
        else:
            print(f"\nUnknown command: {command}\n")

    def _ask_for_quit_confirmation(self) -> bool:
        while True:
            answer = input("Do you want to quit Exponent shell? [y/N] ").strip().lower()

            if answer in {"y", "yes"}:
                return True
            elif answer in {"n", "no", ""}:
                return False

    def _toggle_autorun(self) -> None:
        if self.chat.toggle_autorun():
            print("\nAutorun mode enabled\n")
        else:
            print("\nAutorun mode disabled\n")

    def _switch_chat_to_web(self) -> None:
        url = self.chat.url()
        print(f"\nThis chat has been moved to {url}\n")

        if not inside_ssh_session():
            click.launch(url)

        while True:
            input()

    def _confirm_execution(self) -> None:
        render(
            [
                "\r",
                move_cursor_up_seq(1),
                clear_line_seq(),
            ]
        )

        self._ensure_connected()
        self.stream_fut = self._run_coroutine(self.chat.send_confirmation())

    def _send_initial_prompt(self) -> None:
        if self.prompt is not None:
            self._send_prompt(self.prompt)

    def _send_prompt(self, text: str) -> None:
        self._ensure_connected()
        self.stream_fut = self._run_coroutine(self.chat.send_prompt(text))

    def _wait_for_stream_completion(self) -> None:
        if self.stream_fut is not None:
            self.stream_fut.result()
            self.stream_fut = None

    def _run_coroutine(self, coro: Coroutine[Any, Any, Any]) -> Future[Any]:
        return asyncio.run_coroutine_threadsafe(coro, self.loop)


class SlashCommandCompleter(Completer):
    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        text = document.text
        if text.startswith("/"):
            for command in SLASH_COMMANDS:
                if command.startswith(text):
                    yield Completion(command, start_position=-len(text))


class Block:
    def __init__(self, event_uuid: str) -> None:
        self.event_uuid = event_uuid
        self.spinner = None
        self.block_row_offset = 0
        self.console = Console()
        self.theme = "monokai"

    def highlight_code(self, code: str, lang: str, line_numbers: bool = True) -> str:
        syntax = Syntax(
            code,
            lang,
            theme=self.theme,
            line_numbers=line_numbers,
            word_wrap=True,
        )

        with self.console.capture() as capture:
            self.console.print(syntax)

        return capture.get()


class NullBlock(Block):
    def __init__(self) -> None:
        super().__init__("")


class MessageBlock(Block):
    def __init__(self, event_uuid: str) -> None:
        super().__init__(event_uuid)
        self.content_len = 0

    def render_message(self, message: str) -> list[Any]:
        message = message.strip()
        chunk = message[self.content_len :]
        self.content_len = len(message)

        return [chunk]


class CodeBlock(Block):
    def __init__(self, event_uuid: str, lang: str) -> None:
        super().__init__(event_uuid)
        self.lang = lang
        self.line_count = 0

    def render_code(
        self,
        code: str,
    ) -> list[Any]:
        seqs: list[Any] = []

        code = pad_left(code.strip(), "  ")
        code = self.highlight_code(code, self.lang, line_numbers=False)
        lines = code.split("\n")
        lines = lines[0 : len(lines) - 1]
        new_line_count = len(lines)

        if self.line_count > 0:
            seqs.append([move_cursor_up_seq(1), "\r"])
            lines = lines[self.line_count - 1 :]

        lines = [line + "\n" for line in lines]
        seqs.append(lines)
        self.line_count = new_line_count

        return seqs


class FileWriteBlock(Block):
    def __init__(self, event_uuid: str, lang: str) -> None:
        super().__init__(event_uuid)
        self.lang = lang
        self.line_count = 0

    def render_content(self, content: str) -> list[Any]:
        seqs: list[Any] = []
        code = pad_left(content.strip(), "  ")
        code = self.highlight_code(code, "rust" or self.lang, line_numbers=False)
        lines = code.split("\n")
        lines = lines[0 : len(lines) - 1]
        new_line_count = len(lines)

        if self.line_count > 0:
            seqs.append([move_cursor_up_seq(1), "\r"])
            lines = lines[self.line_count - 1 :]

        lines = [line + "\n" for line in lines]
        seqs.append(lines)
        self.line_count = new_line_count

        return seqs


class CommandBlock(Block):
    def __init__(self, event_uuid: str, type: str, lang: str) -> None:
        super().__init__(event_uuid)
        self.type = type
        self.lang = lang
        self.line_count = 0

    def render_data(self, data: dict[str, Any]) -> list[Any]:
        seqs: list[Any] = []
        paths = self.render_paths(data["filePath"])
        new_line_count = len(paths)

        if self.line_count > 0:
            seqs.append([move_cursor_up_seq(1), "\r"])
            paths = paths[self.line_count - 1 :]

        seqs.append(paths)
        self.line_count = new_line_count

        return seqs

    def render_paths(self, paths: str) -> list[Any]:
        lines = pad_left(paths.strip(), "  ").split("\n")

        return [[block_body_bg_seq(), erase_line_seq(), line, "\n"] for line in lines]


def text_line_count(text: str) -> int:
    cols = get_terminal_width()
    count = 0

    for line in text.split("\n"):
        count += len(line) // cols

        if len(line) == 0 or len(line) % cols > 0:
            count += 1

    return count


def get_terminal_width() -> int:
    cols, _ = shutil.get_terminal_size()
    return cols


def pad_left(text: str, padding: str) -> str:
    return "\n".join([padding + line for line in text.strip().split("\n")])


def fg_color_seq(c: int) -> str:
    return f"\x1b[{30 + c}m"


def block_header_bg_seq() -> str:
    return "\x1b[48;2;29;30;24m"


def block_header_fg_seq() -> str:
    return "\x1b[38;5;246m"


def block_body_bg_seq() -> str:
    return "\x1b[48;5;235m"


def erase_line_seq() -> str:
    return "\x1b[2K"


def erase_display_seq() -> str:
    return "\x1b[0J"


def reset_attrs_seq() -> str:
    return "\x1b[0m"


def clear_line_seq() -> str:
    return f"\r{reset_attrs_seq()}{erase_line_seq()}"


def move_cursor_up_seq(n: int) -> str:
    if n > 0:
        return f"\x1b[{n}A"
    else:
        return ""


def move_cursor_down_seq(n: int) -> str:
    if n > 0:
        return f"\x1b[{n}B"
    else:
        return ""


def render(seqs: str | list[Any] | None) -> int:
    text, newline_count = collect(seqs)
    print(text, end="")
    sys.stdout.flush()

    return newline_count


def collect(seqs: str | list[Any] | None) -> tuple[str, int]:
    if seqs is None:
        return ("", 0)

    if isinstance(seqs, str):
        return (seqs, seqs.count("\n"))

    if isinstance(seqs, tuple):
        return seqs

    assert isinstance(seqs, list)

    text = ""
    newline_count = 0

    for seq in seqs:
        t, c = collect(seq)
        text += t
        newline_count += c

    return (text, newline_count)
